div.hololayout {
    display: flex;
    align-items: center;
    margin: 0;
}

div.holoframe {
	width: 75%;
}

div.holowell {
    display: flex;
    align-items: center;
    margin: 0;
}

form.holoform {
    background-color: #fafafa;
    border-radius: 5px;
    overflow: hidden;
	padding-left: 0.8em;
    padding-right: 0.8em;
    padding-top: 0.4em;
    padding-bottom: 0.4em;
}

div.holowidgets {
    padding-right: 0;
	width: 25%;
}

div.holoslider {
    min-height: 0 !important;
    height: 0.8em;
    width: 60%;
}

div.holoformgroup {
    padding-top: 0.5em;
    margin-bottom: 0.5em;
}

div.hologroup {
    padding-left: 0;
    padding-right: 0.8em;
    width: 50%;
}

.holoselect {
    width: 92%;
    margin-left: 0;
    margin-right: 0;
}

.holotext {
    width: 100%;
    padding-left:  0.5em;
    padding-right: 0;
}

.holowidgets .ui-resizable-se {
	visibility: hidden
}

.holoframe > .ui-resizable-se {
	visibility: hidden
}

.holowidgets .ui-resizable-s {
	visibility: hidden
}
